summary:
    Short description summarizing the story
story:
    As a user I want to do this and that
    so that I can achieve this.
example:
    - One example is worth thousand words.
    - Of course, there can be more than one.
